# -*- coding:UTF8 -*-
# SPIRALE: A one-time pad cryptosystem to be used as paper & pencil cipher
#          for text with letters, numerals and typographic characters
versionname = "SPIRALE 2.1.3"
# - 2015-07-22 by Ph. Allard (program support by AK-BE)
# - Implementation of the paper: "SPIRALE -- A One-Time-Pad Cipher", 2015
# - To simplify the command line interface
#   - the plaintext is always read from or written to the file plaintext.txt.
#   - the ciphertext is always read from or written to the file ciphertext.txt.
#   So, the user has to enter only the cryptographic parameters (always five):
#   Mandatory: - key 1 and key 2 are for permutations of the working alphabet.
#              - key 3 and key 4 are for generating the long key.
#   Optionally: The name of the file with the working alphabet
#               (by default the given file alpha_L.txt is used which contains "ABCD...XYZ")
# Sample usage:
#   copy alpha_LNT.txt  alpha.txt
#   copy plaintext_LNT.txt  plaintext.txt
#   python Spirale.py -e aaaaaaa bbbbbbb ccccccc ddddddd -a alpha -v 

#----------------------------------------- Initialization -------------------------------

# Initialization of variables
lenkey = 7  # length of key (same for all 4 keys; global variable, also used WITHIN defined functions)

# Get and define options (command line parameters)
from optparse import OptionParser
parser = OptionParser(usage = "usage: %prog -e OR -d key1 key2 key3 key4  [-a alpha]  [-v]\n" + versionname +
" -- OTP cryptosystem for paper&pencil usage processing text with letters, numerals and typographic characters according to selected alphabet.")
parser.add_option("-e", "--encrypt", help='Encrypt plaintext.txt to ciphertext.txt.', action="store", type="str", nargs=4)
parser.add_option("-d", "--decrypt", help='Decrypt ciphertext.txt to plaintext.txt.', action="store", type="str", nargs=4)
parser.add_option("-a", "--alphabet", help='Read alphabet to be used (from alpha.txt).', action="store", type="str", nargs=1, default="alpha_L")
parser.add_option("-v", "--verbose", help='Print and store some intermediate results.', action='store_true')
(options, args) = parser.parse_args()

## Interpret/check options, exit if wrong, set according global variables
## (Usage is shown, if options are used incorrectly.)
if options.verbose:
    T_verbose=1
else:
    T_verbose=0
if options.encrypt and options.decrypt:
    parser.error("Options -e and -d are mutually exclusive.")
if not(options.encrypt or options.decrypt): 
    parser.error("Option (either -e or -d) is missing.")

# Set keys for -e or -d
if options.encrypt:
  (K1, K2, K3, K4) = options.encrypt[0:4]
if options.decrypt:
  (K1, K2, K3, K4) = options.decrypt[0:4]
upK1 = K1.upper()	# key in upper case
upK2 = K2.upper()	# key in upper case
upK3 = K3.upper()	# key in upper case
upK4 = K4.upper()	# key in upper case
if len(K1) != lenkey or len(K2) != lenkey or len(K3) != lenkey or len(K4) != lenkey :
    print("Error with option -e or -d: Each of the four key arguments must be", lenkey, "characters long.")
    exit()

# Get the name of the file containing the alphabet (given after -a).
# (If no option -a is set, the default is used: alpha="alpha_L".)
if options.alphabet:
  alpha = options.alphabet

# Set depth of Generalized Fibonacci recurrence: 
depth = 24

  
## helper functions

def writefile(f,s):
    fp = open(f,'w')
    fp.write(s)
    fp.close()

def readfile(f):
   fp = open(f,'r')
   s = fp.read()
   fp.close()
   return s

def str2list(s): return [c for c in s]


# Read-in alphabet to be used into alphabet0
# Here no check of the read-in alphabet is done!
alphabet0 = readfile(alpha+'.txt').upper()      # input of alphabet from text file
modulus = len(alphabet0)

# We expect the alphabet to have a special minimal length of 26
if modulus < 26:
    print("Error: Length of alphabet read from " + alpha+".txt" + " = "+ str(modulus) +". It must at least contain 26 different characters.")
    exit()

print(versionname+":")
if options.encrypt: print("Encryption process from keys ("+upK1+"|"+upK2+"|"+upK3+"|"+upK4+") and "+str(modulus)+"-alphabet ["+alphabet0+"]")
if options.decrypt: print("Decryption process from keys ("+upK1+"|"+upK2+"|"+upK3+"|"+upK4+") and "+str(modulus)+"-alphabet ["+alphabet0+"]")


#-------------- Step 1 : Alphabet Permutations with key 1 and key 2 to generate the ciphering table ------------------------------------

def permutation (upkey) :
    # list of ranks of characters in upkey
    extraction_list = [alphabet0.index(c)+1 for c in upkey]
    # extraction of characters from the right according to upkey
    permuted_list=[]
    pos = 0
    alphabet_list = str2list(alphabet0)
    for i in range(0,len(alphabet_list)):
        pos = (pos-extraction_list[i % lenkey]) % len(alphabet_list)
        permuted_list.append(alphabet_list[pos])
        del alphabet_list[pos]
    permuted_string = ''.join(permuted_list)    # conversion to string
    if T_verbose:
        print("Permutation of alphabet according to following ranks :", extraction_list, " for "+upkey+":")
        print("Permuted alphabet ("+str(modulus)+" characters) = "+permuted_string)
        writefile('PermAlph_'+upkey+'.txt', permuted_string)     # output of sequence in text file
        print ('The permuted alphabet was written to file PermAlph_'+upkey+'.txt')
    return permuted_string  

if T_verbose:
    print("\n1. Generate ciphering table from key 1 and key 2:")

# Check to ensure that each letter from key 1 and key 2 is part of the given alphabet.
# This avoids "ValueError: substring not found" at "extraction_list = [alphabet0.index(c)+1 for c in upkey]" in fct 'permutation'.
for c in upK1:
    if not c in alphabet0:
        print("Error: Key 1: Character " + c + " is not part of the given alphabet ["+alphabet0+"].")
        exit()
for c in upK2:
    if not c in alphabet0:
        print("Error: Key 2: Character " + c + " is not part of the given alphabet ["+alphabet0+"].")
        exit()

alphabet1 = permutation(upK1)       # permuted alphabet for rows in the ciphering table
alphabet2 = permutation(upK2)       # permuted alphabet for columns in the ciphering table
alphabets = [alphabet0,alphabet1,alphabet2]


#-------------- Step 2 : Create long key matrix with key 3 and key 4 to generate the long key -----------------------------------

# 'square'-operator to get the result value derived from ciphering table
# in order to generate the product matrix, the linear recurrence and to do encryption (in step 4a)
def square (c1, c2):
    v1 = alphabet1.index(c1)
    v2 = alphabet2.index(c2)
    v = ( v1+v2 ) % modulus
    return alphabet0[v]

# Inverse operator 'unsquare' derived from ciphering table
# in order to do decryption (in step 4b)
def unsquare (c0,c2):
    v0 = alphabet0.index(c0)
    v2 = alphabet2.index(c2)
    v1 = ( v0 - v2 ) % modulus
    return alphabet1[v1]

if T_verbose:
    print("\n2. Generate long key from key 3 and key 4:")

# Check to ensure that each letter from key 3 and key 4 is part of the given alphabet.
# This avoids "ValueError: substring not found" in fct 'square'/'unsquare'.
for c in upK3:
    if not c in alphabet0:
        print("Error: Key 3: Character " + c + " is not part of the given alphabet ["+alphabet0+"].")
        exit()
for c in upK4:
    if not c in alphabet0:
        print("Error: Key 4: Character " + c + " is not part of the given alphabet ["+alphabet0+"].")
        exit()

# Generation of the product matrix (via key3 square key4)
table = [ [ square(c1,c2) for c2 in upK4 ] for c1 in upK3 ]
if T_verbose:
    print("Product matrix by key 3 and key 4:")
    for row in table: print(row)

# Reading diagonally from the product matrix
longkey_list=[]

for n in range(lenkey):
    p,q = 0,n
    while q >= 0 and p < lenkey:
        longkey_list.append(table[q][p])
        p,q = p+1,q-1
for p in range(1,lenkey):
    q = lenkey-1
    while q >= 0 and p < lenkey:
        longkey_list.append(table[q][p])
        p,q = p+1,q-1

if T_verbose:
    longkey_string = ''.join(longkey_list)
    print("Generated long key ("+str(lenkey*lenkey)+" characters) = "+longkey_string)


#-------------- Step 3 : Keystream Generation -------------------------------

if T_verbose:
    print("\n3. Generate keystream:")

keystream_list = longkey_list  # long key at beginning of keystream list
k = len(keystream_list)        # maximum depth of recurrence
d = depth                      # intermediate depth of recurrence
if k == 2*d:
    print("Error: Long key length k="+ k +" must not be the double of depth d="+ d +". Change one of them (d directly in the code).")
    exit()

if options.encrypt:
    plaintext_string = readfile('plaintext.txt').upper()   # input of plaintext from text file
    lentext = len(plaintext_string)
    if lentext <= k: print("Short plaintext! (length =", lentext, ")")
    # Check to ensure that each letter from read plaintext is part of the given alphabet.
    # This avoids "ValueError: substring not found" in fct 'square'.
    for c in plaintext_string:
        if not c in alphabet0:
            print("Error: Plaintext: Read character " + c + " is not part of the given alphabet ["+alphabet0+"].")
            exit()
        
if options.decrypt:
    ciphertext_string = readfile('ciphertext.txt').upper() # input of ciphertext from text file
    lentext = len(ciphertext_string)
    if lentext <= k: print("Short ciphertext! (length =", lentext, ")")
    # Check to ensure that each letter from read ciphertext is part of the given alphabet.
    # This avoids "ValueError: substring not found" in fct 'unsquare'.
    for c in ciphertext_string:
        if not c in alphabet0:
            print("Error: Ciphertext: Read character " + c + " is not part of the given alphabet ["+alphabet0+"].")
            exit()

# Generating the generalized fibonacci sequence of depths k,d
for n in range(k,lentext):		    # generation of letter n
    keystream_list.append( square(keystream_list[n-k],keystream_list[n-d]) )
    
if T_verbose:
    keystream_string = ''.join(keystream_list)     # conversion to string
    print("Keystream ("+str(lentext)+" characters) = "+keystream_string)
    writefile('keystream.txt',keystream_string)    # output of sequence in text file
    print ('The keystream sequence of characters was written to file keystream.txt.')


#-------------- Step 4a: Encryption -----------------------------------------

if options.encrypt:
    if T_verbose:
        print("\n4. Encryption with OTP:")
    plaintext_list = str2list( plaintext_string)
    ciphertext_list = []
    for n in range(lentext):         # encrypt character n
        ciphertext_list.append( square(plaintext_list[n],keystream_list[n]) )
    ciphertext_string = ''.join(ciphertext_list)	
    writefile('ciphertext.txt', ciphertext_string)
    print ("The generated ciphertext was written to 'ciphertext.txt'.")
    if T_verbose:
        print("Plaintext:  " + plaintext_string)
        print("Ciphertext: " + ciphertext_string)
        print("")


#-------------- Step 4b: Decryption -----------------------------------------

if options.decrypt:
    if T_verbose:
        print("\n4. Decryption with OTP:")
    ciphertext_list = str2list( ciphertext_string )
    plaintext_list = []
    for n in range(lentext):         # decrypt character n
        plaintext_list.append( unsquare(ciphertext_list[n],keystream_list[n]) )
    plaintext_string = ''.join(plaintext_list)	
    writefile('plaintext.txt', plaintext_string)
    print ("The generated plaintext was written to 'plaintext.txt'.")
    if T_verbose:
        print("Ciphertext: " + ciphertext_string)
        print("Plaintext:  " + plaintext_string)
        print("")

#-------------------------------------------------------------------------------------------------------------------

